/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-resource"

#include "slurp-resource.h"

typedef struct
{
    GFile *source;
    GFile *target;
    gboolean overwrite;

    gint status;
    guint64 total;
    guint64 downloaded;
    guint progress;
} SlurpResourcePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpResource, slurp_resource, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SOURCE,
    PROP_TARGET,
    PROP_OVERWRITE,
    PROP_TOTAL,
    PROP_DOWNLOADED,
    PROP_PROGRESS,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES];

enum {
    STARTED,
    FINISHED,
    CANCELLED,
    PROGRESS,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

static void
slurp_resource_finalize (GObject *object)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_clear_pointer (&priv->source, g_object_unref);
    g_clear_pointer (&priv->target, g_object_unref);

    G_OBJECT_CLASS (slurp_resource_parent_class)->finalize (object);
}

static void
slurp_resource_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            g_value_set_object (value, priv->source);
            break;

        case PROP_TARGET:
            g_value_set_object (value, priv->target);
            break;

        case PROP_OVERWRITE:
            g_value_set_boolean (value, priv->overwrite);
            break;

        case PROP_TOTAL:
            g_value_set_uint64 (value, priv->total);
            break;

        case PROP_DOWNLOADED:
            g_value_set_uint64 (value, priv->downloaded);
            break;

        case PROP_PROGRESS:
            g_value_set_uint (value, priv->progress);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            priv->source = g_value_dup_object (value);
            break;

        case PROP_TARGET:
            priv->target = g_value_dup_object (value);
            break;

        case PROP_OVERWRITE:
            priv->overwrite = g_value_get_boolean (value);
            break;

        case PROP_TOTAL:
            priv->total = g_value_get_uint64 (value);
            break;

        case PROP_DOWNLOADED:
            priv->downloaded = g_value_get_uint64 (value);
            break;

        case PROP_PROGRESS:
            priv->progress = g_value_get_uint (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_class_init (SlurpResourceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_resource_finalize;
    object_class->get_property = slurp_resource_get_property;
    object_class->set_property = slurp_resource_set_property;

	/**
	 * SlurpResource:source:
	 *
	 *
	 */
    properties[PROP_SOURCE] =
        g_param_spec_object ("source",
                             "Source",
                             "",
                             G_TYPE_FILE,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:target:
	 *
	 *
	 */
    properties[PROP_TARGET] =
        g_param_spec_object ("target",
                             "Target ",
                             "",
                             G_TYPE_FILE,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:overwrite:
	 *
	 *
     *
     * Recommended: %FALSE
	 */
    properties[PROP_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite",
                              "",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:total:
	 *
	 *
	 */
    properties[PROP_TOTAL] =
        g_param_spec_uint64 ("total",
                             "Total Download Size",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:downloaded:
	 *
	 *
	 */
    properties[PROP_DOWNLOADED] =
        g_param_spec_uint64 ("downloaded",
                             "Downloaded Bytes",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:progress:
	 *
	 *
	 */
    properties[PROP_PROGRESS] =
        g_param_spec_uint ("progress",
                           "Progress",
                           "",
                           0,
                           G_MAXUINT,
                           0,
                           G_PARAM_STATIC_STRINGS | G_PARAM_READABLE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);

    /**
     * SlurpResource::started:
     * @object: the object which received the signal
     *
     * The ::started signal is emitted when the resource begins downloading.
     */
    signals[STARTED] =
        g_signal_new ("started",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    /**
     * SlurpResource::finished:
     * @object: the object which received the signal
     *
     * The ::finished signal is emitted when the downloaded has finished.
     */
    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    /**
     * SlurpResource::cancelled:
     * @object: the object which received the signal
     *
     * The ::cancelled signal is emitted when the
     */
    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    /**
     * SlurpResource::progress:
     * @object: the object which received the signal
     * @total: The total size of the resoure, 0 if no total size is set
     * @downloaded: The number of bytes downloaded
     *
     * The ::progress signal is emitted when the download progress of the
     * resource is :updated.
     */
    signals[PROGRESS] =
        g_signal_new ("progress",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      2,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64);
}

static void
slurp_resource_init (SlurpResource *self)
{
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpResource *
slurp_resource_new (void)
{
    return g_object_new (SLURP_TYPE_RESOURCE, NULL);
}

GFile *
slurp_resource_get_source (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    GValue value = G_VALUE_INIT;
    GObject *retval;

    g_value_init (&value, G_TYPE_OBJECT);
    g_object_get_property (G_OBJECT (self), "source", &value);
    retval = g_value_dup_object (&value);
    g_value_unset (&value);
    return G_FILE (retval);
}

void
slurp_resource_set_source (SlurpResource *self,
                           GFile         *source)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));
    g_return_if_fail (G_IS_FILE (source));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_OBJECT);
    g_value_set_object (&value, source);
    g_object_set_property (G_OBJECT (self), "source", &value);
    g_value_unset (&value);
}

void
slurp_resource_set_source_from_uri (SlurpResource *self,
                                    const gchar   *uri)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;
    GFile *source = NULL;

    source = g_file_new_for_uri (uri);

    g_value_init (&value, G_TYPE_FILE);
    g_value_set_object (&value, source);
    g_object_set_property (G_OBJECT (self), "source", &value);
    g_value_unset (&value);
}

GFile *
slurp_resource_get_target (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    GValue value = G_VALUE_INIT;
    GObject *retval;

    g_value_init (&value, G_TYPE_OBJECT);
    g_object_get_property (G_OBJECT (self), "target", &value);
    retval = g_value_dup_object (&value);
    g_value_unset (&value);
    return G_FILE (retval);
}

void
slurp_resource_set_target (SlurpResource *self,
                           GFile         *target)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));
    g_return_if_fail (G_IS_FILE (target));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_OBJECT);
    g_value_set_object (&value, target);
    g_object_set_property (G_OBJECT (self), "target", &value);
    g_value_unset (&value);
}

void
slurp_resource_set_target_from_uri (SlurpResource *self,
                                    const gchar   *uri)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;
    GFile *target = NULL;

    target = g_file_new_for_uri (uri);

    g_value_init (&value, G_TYPE_FILE);
    g_value_set_object (&value, target);
    g_object_set_property (G_OBJECT (self), "target", &value);
    g_value_unset (&value);
}

gboolean
slurp_resource_get_overwrite (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);

    GValue value = G_VALUE_INIT;
    gboolean retval;

    g_value_init (&value, G_TYPE_BOOLEAN);
    g_object_get_property (G_OBJECT (self), "overwrite", &value);
    retval = g_value_get_boolean (&value);
    g_value_unset (&value);
    return retval;
}

void
slurp_resource_set_overwrite (SlurpResource *self,
                              gboolean       overwrite)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_BOOLEAN);
    g_value_set_boolean (&value, overwrite);
    g_object_set_property (G_OBJECT (self), "overwrite", &value);
    g_value_unset (&value);
}
