
enum {
    SLURP_RESOURCE_ERROR,
    SLURP_RESOURCE_ERROR_SOURCE,
    SLURP_RESOURCE_ERROR_SOURCE_MISSING,
    SLURP_RESOURCE_ERROR_SOURCE_INVALID_URI,
    SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
    SLURP_RESOURCE_ERROR_TARGET,
    SLURP_RESOURCE_ERROR_TARGET_MISSING,
    SLURP_RESOURCE_ERROR_TARGET_INVALID_URI,
    SLURP_RESOURCE_ERROR_TARGET_INVALID_SCHEME,
    SLURP_RESOURCE_ERROR_WRITABLE
};

/**
 * slurp_resource_is_support_scheme:
 * @scheme: a string representing the scheme
 *
 * Checks whether a scheme is supported by slurp. This function does no
 * blocking IO.
 *
 * Returns: %TRUE if scheme is supported, or %FALSE if not supported
 */
gboolean
slurp_resource_is_support_scheme (SlurpResource *self,
                                  const gchar   *scheme)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return g_hash_table_contains (priv->schemes, scheme);
}

gboolean
slurp_resource_is_valid_source (SlurpResource  *self,
                                const gchar    *uri,
                                GCancellable   *cancellable,
                                GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    GFile *file;

    if (g_strcmp0 (uri, "") == 0)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
                     "Source can not be empty");
        goto failed;
    }

    // Check URI

    file = g_file_new_for_uri (uri);

    if (!slurp_resource_is_support_scheme (self, g_file_get_uri_scheme (file)))
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
                     "Source is not a support scheme type.");
    }

    g_object_unref (file);
    return TRUE;

failed:
    return FALSE;
}

gboolean
slurp_resource_is_valid_target (SlurpResource  *self,
                                const gchar    *uri,
                                GCancellable   *cancellable,
                                GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_TARGET_MISSING,
                     "Target is not set");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_writable (SlurpResource  *self,
                            const gchar    *uri,
                            GCancellable   *cancellable,
                            GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_TARGET_MISSING,
                     "Target is not set");
        return FALSE;
    }

    if (!priv->overwrite && g_file_query_exists (priv->file, cancellable))
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_WRITABLE,
                     "Target is not writable");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_valid (SlurpResource  *self,
                         GCancellable   *cancellable,
                         GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_debug ("Source: %s", priv->source);
    g_debug ("Target: %s", priv->target);

    return TRUE;

/*
    GError *internal_error = NULL;

    if (!slurp_resource_is_valid_source (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_valid_target (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_writable (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    return TRUE;

out:
    g_clear_error (&internal_error);
    return FALSE;
*/
}
