/* slurp-download.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource downloader using the libsoup library
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GCallback, #GCancellable, #GFile,
 *            #GInputStream
 *
 * Asynchronous resource downloader using the libsoup library which offers
 * addded features over using plain `soup_*` calls.
 **/

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    SoupSession *session;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SESSION,
    PROP_USER_AGENT,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    SLURP_DOWNLOAD_ERROR,
    SLURP_DOWNLOAD_ERROR_NO_SESSION
};

static void
_download_resource_async_read_cb (GObject      *object,
                                  GAsyncResult *result,
                                  gpointer      user_data)
{

}

static void
_download_resource_async_cb (GObject      *object,
                             GAsyncResult *result,
                             gpointer      user_data)
{

}

static void
_download_resource_in_thread_func (GTask        *task,
                                   gpointer      source,
                                   gpointer      task_data,
                                   GCancellable *cancellable)
{
    SlurpDownload *self = SLURP_DOWNLOAD (source);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);
    SlurpDownloadData *data = (SlurpDownloadData *) task_data;
    GMainContext *download_context;
    SoupRequest *request;
    GError *error = NULL;

    g_return_if_fail (G_IS_TASK (task));
    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (g_task_is_valid (task, self));

    if (g_task_return_error_if_cancelled (task))
        return;

    download_context = g_main_context_new ();
    g_main_context_push_thread_default (download_context);

    data->loop = g_main_loop_new (download_context, FALSE);

    // TODO: Error if no session
    if (!priv->session)
    {
        error = g_error_new (SLURP_DOWNLOAD_ERROR,
                             SLURP_DOWNLOAD_ERROR_NO_SESSION,
                             "No libsoup session.");

        g_debug ("Error: %s", error->message);
        g_task_return_error (task, error);
        goto out;
    }

    request = soup_session_request (priv->session,
                                    slurp_resource_get_source (data->resource),
                                    &error);

    if (!request)
    {
        g_debug ("Error: %s", error->message);
        g_task_return_error (task, error);
        goto out;
    }
/*
    soup_request_send_async (SOUP_REQUEST(request),
                             cancellable,
                             _download_resource_async_cb,
                             data);
*/
    g_main_loop_run (data->loop);
    g_task_return_boolean (task, TRUE);

out:
    g_main_context_pop_thread_default (download_context);
    g_main_context_unref (download_context);
}

static void
_download_resource_async (SlurpDownload       *self,
                          SlurpResource       *resource,
                          GCancellable        *cancellable,
                          GAsyncReadyCallback  callback,
                          gpointer             user_data)
{
    GTask *task = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    data = slurp_download_data_new ();
    data->resource = g_object_ref (resource);

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, _download_resource_async);
    g_task_set_task_data (task, data, (GDestroyNotify) g_object_unref);
    g_task_set_return_on_cancel (task, TRUE);
    g_task_run_in_thread (task, _download_resource_in_thread_func);
    g_object_unref (task);

}

static gboolean
_download_resource_finish (SlurpDownload  *self,
                           GAsyncResult   *result,
                           GError        **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (self), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, self), FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);
    g_clear_pointer (&priv->session, g_object_unref);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            g_value_set_object (value, priv->session);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            priv->session = g_value_dup_object (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:session:
	 *
	 * A SoupSession which Slurp can use to download resources.
     *
     * Note: Make sure to ref your #SoupSession before passing it.
     *
     * g_object_new():
     * |[
     * g_object_new (SLURP_TYPE_DOWNLOAD,
     *               "session", g_object_ref (session),
     *               NULL);
     * ]|
     *
     * slurp_download_new_with_session():
     * |[
     * SlurpDownload *download;
     * download = slurp_download_new_with_session (g_object_ref (session));
     * ]|
	 */
    properties[PROP_SESSION] =
        g_param_spec_object ("session",
                             "A SoupSession",
                             "A SoupSession which Slurp can use to download resources.",
                             SOUP_TYPE_SESSION,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    // TODO: Add new user-agent if session is passed
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "slurp-download",
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

}

static void
slurp_download_init (SlurpDownload *self)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    if (!priv->session)
    {
        priv->session =
            soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                           SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                           SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                           SOUP_SESSION_TIMEOUT, 60,
                                           SOUP_SESSION_IDLE_TIMEOUT, 60,
                                           NULL);
    }
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}

SlurpDownload *
slurp_download_new_with_session (SoupSession *session)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD,
                         "session", session,
                         NULL);
}

void
slurp_download_resource_async (SlurpDownload       *self,
                               SlurpResource       *resource,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  callback,
                               gpointer             user_data)
{
    _download_resource_async (self, resource, cancellable, callback, user_data);
}

gboolean
slurp_download_resource_finish (SlurpDownload  *self,
                                GAsyncResult   *result,
                                GError        **error)
{
    return _download_resource_finish (self, result, error);
}
